;; Generated package description from cus-abbrev.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "cus-abbrev" "0.1.0.20250305.64248" "Easy Customization interface for Abbrevs" '((emacs "29.1")) :commit "1fe96883a9cfd3c2cc6381049d95e6330907f140" :authors '(("Mauro Aranda" . "maurooaranda@gmail.com")) :maintainer '("Mauro Aranda" . "maurooaranda@gmail.com") :keywords '("convenience") :url "https://gitlab.com/mauroaranda/cus-abbrev")
